﻿using System;
 
 
using System.Text;

using System.Collections;

using STMBench7.Annotations;
using STMBench7.Operations;
using System.Threading;
using STMBench7.Core;

namespace STMBench7
{

	/**
	 * A single thread of the STMBench7 benchmark. Executes operations
	 * assigned to it one by one, randomly choosing the next
	 * operation and respecting the expected ratios of operations'
	 * counts.
	 */
	[NonAtomic]
	public class BenchThread /*: Thread*/ {

        //added in port because Thread is final
        private Thread realThread;
        public void Start()
        {
            realThread.Start();
        }
        public void Join()
        {
            realThread.Join();
        }
        public void Sleep(int ms)
        {
            Thread.Sleep(ms);
        }
        //


        private readonly STMBench7.Locking.MGLockingOperationExecutor.AssemblyLocksAcquired assemblyLocksAcquired = new STMBench7.Locking.MGLockingOperationExecutor.AssemblyLocksAcquired();



	    protected volatile bool stop = false;
	    protected double[] operationCDF;
	    protected OperationExecutor[] operations;
	    protected readonly short myThreadNum;
	    
	    public int[] successfulOperations, failedOperations;
	    public int[,] operationsTTC, operationsHighTTCLog;
	    
	    public class ReplayLogEntry : IComparable {
	    	public readonly short threadNum;
	    	public readonly int timestamp, result;
	    	public readonly bool failed;
	    	
	    	public ReplayLogEntry(int myThreadNum, int timestamp, int result, bool failed) {
	    		this.threadNum = (short)myThreadNum;
	    		this.timestamp = timestamp;
	    		this.result = result;
	    		this.failed = failed;
	    	}

			public int CompareTo(Object entry) {
                ReplayLogEntry e = (ReplayLogEntry)entry;
				return timestamp - e.timestamp;
			}
	    }
	    
	    public ArrayList/*<ReplayLogEntry>*/ replayLog;
	    
	    public BenchThread(Setup setup, double[] operationCDF, short myThreadNum)
        {
            realThread = new Thread(new ThreadStart(this.Run));

	    	this.operationCDF = operationCDF;

            int numOfOperations = OperationId.values().Length;
	    	operationsTTC = new int[numOfOperations,Parameters.MAX_LOW_TTC + 1];
	    	operationsHighTTCLog = new int[numOfOperations,Parameters.HIGH_TTC_ENTRIES];
	    	successfulOperations = new int[numOfOperations];
	    	failedOperations = new int[numOfOperations];
	    	operations = new OperationExecutor[numOfOperations];
	    	this.myThreadNum = myThreadNum;
	    	
	    	CreateOperations(setup);
	    	
	    	if(Parameters.sequentialReplayEnabled) replayLog = new ArrayList/*<ReplayLogEntry>*/();
	    }

	    protected BenchThread(Setup setup, double[] operationCDF) {
	    	this.operationCDF = operationCDF;
	    	operations = new OperationExecutor[OperationId.values().Length];
	    	CreateOperations(setup);
	    	myThreadNum = 0;
	    }
	    
	    public virtual void Run() {
	    	while(! stop) {
	    		int operationNumber = GetNextOperationNumber();
				OperationExecutor currentExecutor = operations[operationNumber];
				int result = 0;
				bool failed = false;
				
	    		try {    	   		
	    			//long startTime = System.currentTimeMillis();
                    DateTime startTime = DateTime.Now;

                    result = currentExecutor.Execute();

	    			//long endTime = System.currentTimeMillis();
                    DateTime endTime = DateTime.Now;


	    			successfulOperations[operationNumber]++;


                    //tick= 100nanos
                    int ttc = (int)((endTime.Ticks - startTime.Ticks) / 10000.0);
	    			if(ttc <= Parameters.MAX_LOW_TTC) operationsTTC[operationNumber,ttc]++;
	    			else {
	    				double logHighTtc = (Math.Log(ttc) - Math.Log(Parameters.MAX_LOW_TTC + 1)) / 
	    					Math.Log(Parameters.HIGH_TTC_LOG_BASE);
	    				int intLogHighTtc = Math.Min( (int)logHighTtc, Parameters.HIGH_TTC_ENTRIES - 1);
	    				operationsHighTTCLog[operationNumber,intLogHighTtc]++;
	    			}
	    		}
	    		catch(OperationFailedException ex) {

                    //Console.WriteLine(currentExecutor.GetLastOperationTimestamp());
                    //Console.WriteLine(ex.ToString());
                    //Console.WriteLine(ex.StackTrace);

                    failedOperations[operationNumber]++;
	    			failed = true;
	    		}
	    		
				if(Parameters.sequentialReplayEnabled) {

					ReplayLogEntry newEntry = 
						new ReplayLogEntry(myThreadNum, currentExecutor.GetLastOperationTimestamp(), result, failed);

					replayLog.Add(newEntry);
				}


	    	}
	    	Console.Error.WriteLine("Thread #" + myThreadNum + " finished.");
	    }

	    public void StopThread() {
	    	stop = true;
	    }
	    
	    protected void CreateOperations(Setup setup) {
	    	foreach(OperationId operationDescr in OperationId.values()) {
	    		Type /*Class<? extends Operation>*/ operationClass = operationDescr.GetOperationClass();
	    		int operationIndex = operationDescr.ordinal();
	    		
	    		try {
                    //ConstructorInfo operationConstructor = operationClass.GetConstructor(new Type[]{typeof(Setup)});
	    			//Operation operation = (Operation)operationConstructor.Invoke(new Object[]{setup});
					
					Operation operation;
					
					if(operationClass==typeof(Operation6)) operation = new Operation6(setup);
					else if(operationClass==typeof(Operation7)) operation = new Operation7(setup);
					else if(operationClass==typeof(Operation8)) operation = new Operation8(setup);
					else if(operationClass==typeof(Operation9)) operation = new Operation9(setup);
					else if(operationClass==typeof(Operation10)) operation = new Operation10(setup);
					else if(operationClass==typeof(Operation11)) operation = new Operation11(setup);
					else if(operationClass==typeof(Operation12)) operation = new Operation12(setup);
					else if(operationClass==typeof(Operation13)) operation = new Operation13(setup);
					else if(operationClass==typeof(Operation14)) operation = new Operation14(setup);
					else if(operationClass==typeof(Operation15)) operation = new Operation15(setup);
					else if(operationClass==typeof(Query1)) operation = new Query1(setup);
					else if(operationClass==typeof(Query2)) operation = new Query2(setup);
					else if(operationClass==typeof(Query3)) operation = new Query3(setup);
					else if(operationClass==typeof(Query4)) operation = new Query4(setup);
					else if(operationClass==typeof(Query5)) operation = new Query5(setup);
					else if(operationClass==typeof(Query6)) operation = new Query6(setup);
					else if(operationClass==typeof(Query7)) operation = new Query7(setup);
					else if(operationClass==typeof(ShortTraversal1)) operation = new ShortTraversal1(setup);
					else if(operationClass==typeof(ShortTraversal2)) operation = new ShortTraversal2(setup);
					else if(operationClass==typeof(ShortTraversal6)) operation = new ShortTraversal6(setup);
					else if(operationClass==typeof(ShortTraversal7)) operation = new ShortTraversal7(setup);
					else if(operationClass==typeof(ShortTraversal8)) operation = new ShortTraversal8(setup);
					else if(operationClass==typeof(ShortTraversal9)) operation = new ShortTraversal9(setup);
					else if(operationClass==typeof(ShortTraversal10)) operation = new ShortTraversal10(setup);
					else if(operationClass==typeof(StructuralModification1)) operation = new StructuralModification1(setup);
					else if(operationClass==typeof(StructuralModification2)) operation = new StructuralModification2(setup);
					else if(operationClass==typeof(StructuralModification3)) operation = new StructuralModification3(setup);
					else if(operationClass==typeof(StructuralModification4)) operation = new StructuralModification4(setup);
					else if(operationClass==typeof(StructuralModification5)) operation = new StructuralModification5(setup);
					else if(operationClass==typeof(StructuralModification6)) operation = new StructuralModification6(setup);
					else if(operationClass==typeof(StructuralModification7)) operation = new StructuralModification7(setup);
					else if(operationClass==typeof(StructuralModification8)) operation = new StructuralModification8(setup);
					else if(operationClass==typeof(Traversal1)) operation = new Traversal1(setup);
					else if(operationClass==typeof(Traversal2a)) operation = new Traversal2a(setup);
					else if(operationClass==typeof(Traversal2b)) operation = new Traversal2b(setup);
					else if(operationClass==typeof(Traversal2c)) operation = new Traversal2c(setup);
					else if(operationClass==typeof(Traversal3a)) operation = new Traversal3a(setup);
					else if(operationClass==typeof(Traversal3b)) operation = new Traversal3b(setup);
					else if(operationClass==typeof(Traversal3c)) operation = new Traversal3c(setup);
					else if(operationClass==typeof(Traversal4)) operation = new Traversal4(setup);
					else if(operationClass==typeof(Traversal5)) operation = new Traversal5(setup);
					else if(operationClass==typeof(Traversal6)) operation = new Traversal6(setup);
					else if(operationClass==typeof(Traversal7)) operation = new Traversal7(setup);
					else if(operationClass==typeof(Traversal8)) operation = new Traversal8(setup);
					else if(operationClass==typeof(Traversal9)) operation = new Traversal9(setup);
					else{
						Console.WriteLine("unexpected operation class:");
						Console.WriteLine(operationClass);
						throw new Exception();
					}

	    	   		operations[operationIndex] = 
	    	   			OperationExecutorFactory.instance.CreateOperationExecutor(operation);

	    			if(!operation.GetOperationId().GetOperationClass().Equals(operationClass))
                        throw new RuntimeError("assertion failed");
                    
	    		}
	    		catch(Exception e) {
	    			throw new RuntimeError("Error while creating operation " + operationDescr, e);
	    		}
	    	}
	    }
	    
	    protected int GetNextOperationNumber() {
			double whichOperation = ThreadRandom.NextDouble();
			int operationNumber = 0;
	 		while(whichOperation >= operationCDF[operationNumber]) operationNumber++;
	 		return operationNumber;
	    }
	}

}
