﻿using System;
 
 
using System.Text;

using STMBench7.Annotations;
using STMBench7.Correctness.Invariants;
using STMBench7.Core;
using STMBench7.Correctness.Opacity;
using STMBench7.Backend;
using STMBench7.Locking;

using System.Threading;
using STMBench7.Compatability;
using System.Collections;
using STMBench7.Operations;

namespace BartokSTMExtensions
{
    class NoLog : Attribute { }
    class NoLogTrans : Attribute { }
}

namespace STMBench7
{
	/**
	 * STMBench7 benchmark, the main program. 
	 * Run with argument "-h" or "--help" to see the syntax.
	 * 
	 * TODO: The class got too large and needs some careful refactoring.
	 * TODO: An XML output of the benchmark results would be helpful.
	 */
	[NonAtomic]
	public class Benchmark {

        [BartokSTMExtensions.NoLogTrans]
        public static void NonAtomicPrint(String s)
        {
            Console.WriteLine(s);
        }

        public static bool TransactionSafeEquals(Object a, Object b)
        {
            if(a is IndexKey && b is IndexKey)
            {
                IndexKey aa = (IndexKey)a;
                IndexKey bb = (IndexKey)b;
                return aa.Equals(bb);
            }
            if (a is Thread && b is Thread)
            {
                Thread aa = (Thread)a;
                Thread bb = (Thread)b;
                return aa.Equals(bb);
            }
            if (a is Type && b is Type)
            {
                Type aa = (Type)a;
                Type bb = (Type)b;
                return aa.Equals(bb);
            }
            if (a is String && b is String)
            {
                String aa = (String)a;
                String bb = (String)b;
                return aa.Equals(bb);
            }
            return a == b;
        }


		public static readonly String VERSION = "12.22.2009";
		
		class BenchmarkParametersException : Exception {

            public BenchmarkParametersException(String message, Exception cause)
                : base(message, cause)
            {
			}

            public BenchmarkParametersException(String message)
                : base(message)
            {
			}

            public BenchmarkParametersException()
                : base("")
            {
			}
		}
		
	    public static void Main(String[] args) {

	    	Benchmark benchmark = null;

	    	try {
	    		benchmark = new Benchmark(args);
	    	}
	    	catch(BenchmarkParametersException e) {
                
	    		Console.Error.WriteLine(e.Message);
	    		Environment.Exit(1);
	    	}

	    //	benchmark.CheckInvariants(true);
	    	benchmark.CreateInitialClone();

            GC.Collect();

	    	benchmark.Start();
	    //	benchmark.CheckInvariants(false);
	    	benchmark.CheckOpacity();
	    	benchmark.ShowTTCHistograms();
	    	benchmark.ShowStats();
	    }

	    private bool printTTCHistograms = false;
	    private double[] operationCDF;
	    private BenchThread[] threads;
	    private Setup setup, setupClone;
	    private double elapsedTime;
	    
	    private Benchmark(String[] args) {
	    	PrintHeader();
	    	
	    	ParseCommandLineParameters(args);
	    	PrintRunTimeParametersInformation();
	    	
	    	GenerateOperationCDF();
	    	SetupStructures();
		}
	    
	    private void PrintHeader() {
	    	String header =
	    		"The STMBench7 Benchmark (C# Version)\n" +
                "A benchmark for comparing synchronization techniques\n" +
                "Ported from Java version by Ryan Macnak and Jason Ward\n" +
                " on behalf of S3 Lab, Purdue University\n" +
	    		"Version: " + VERSION + "\n" +
	    		"More information: http://lpd.epfl.ch/site/research/tmeval\n" +
	    		"Copyright (C) 2006-2008 LPD, I&C, EPFL (http://lpd.epfl.ch)\n" +
	    		"Implemented by Michal Kapalka (http://kapalka.eu)";
	    	
	    	PrintLine('=');
	    	Console.WriteLine(header);
            
	    	PrintLine('=');
	    	Console.WriteLine();
	    }

		private void ParseCommandLineParameters(String[] args) {
	    	int argNumber = 0;
	   		String workload = null, synchType = null, stmInitializerClassName = null;
	   	    SynchMethodInitializer synchMethodInitializer;
	   	 
	   		while(argNumber < args.Length) {
	    		String currentArg = args[argNumber++];
	    		
	    		try {
	    			if(currentArg.Equals("--help") || currentArg.Equals("-h")) {
	    				PrintSyntax();
	    				throw new BenchmarkParametersException();
	    			}
	    			else if(currentArg.Equals("--no-traversals")) Parameters.longTraversalsEnabled = false;
	    			else if(currentArg.Equals("--no-sms")) Parameters.structureModificationEnabled = false;
	    			else if(currentArg.Equals("--ttc-histograms")) printTTCHistograms = true;
	    			else if(currentArg.Equals("--seq-replay")) Parameters.sequentialReplayEnabled = true;
	    			else if(currentArg.Equals("--")) {
	    				Parameters.stmCommandLineParameters = new String[args.Length - argNumber];
	    				int stmArgNum = 0;
	    				while(argNumber < args.Length) 
	    					Parameters.stmCommandLineParameters[stmArgNum++] = args[argNumber++];
	    				break;
	    			}
	    			else {
	    				String optionValue = args[argNumber++];
	    				if(currentArg.Equals("-t")) Parameters.numThreads = int.Parse(optionValue);
	    				else if(currentArg.Equals("-l")) Parameters.numSeconds = int.Parse(optionValue);
	    				else if(currentArg.Equals("-w")) workload = optionValue;
	    				else if(currentArg.Equals("-g")) synchType = optionValue;
	    				else if(currentArg.Equals("-s")) stmInitializerClassName = optionValue;
	    				else throw new BenchmarkParametersException("Invalid option: " + currentArg);
	    			}
	    		}
	    		catch(IndexOutOfRangeException) {
	    			throw new BenchmarkParametersException("Missing value after option: " + currentArg);
	    		}
	    		catch(FormatException) {
	    			throw new BenchmarkParametersException("Number expected after option: " + currentArg);
	    		}
	    	}
	    	
	    	if(workload != null) {
	    		if(workload.Equals("r")) 
	    			Parameters.workloadType = WorkloadType.READ_DOMINATED;
	    		else if(workload.Equals("rw")) 
	    			Parameters.workloadType = WorkloadType.READ_WRITE;
	    		else if(workload.Equals("w")) 
	    			Parameters.workloadType = WorkloadType.WRITE_DOMINATED;
                else if(workload.Equals("stm"))
                    Parameters.workloadType = WorkloadType.STM_BIASED;
	    		else throw new BenchmarkParametersException("Invalid workload type: " + workload);
	    	}
	    	
	    	if(synchType != null) {
	    		if(synchType.Equals("coarse")) 
	    			Parameters.synchronizationType = Parameters.SynchronizationType.LOCK_COARSE;
	    		else if(synchType.Equals("medium")) 
	    			Parameters.synchronizationType = Parameters.SynchronizationType.LOCK_MEDIUM;
	    		else if(synchType.Equals("fine")) 
	    			Parameters.synchronizationType = Parameters.SynchronizationType.LOCK_FINE;
	    		else if(synchType.Equals("none")) 
	    			Parameters.synchronizationType = Parameters.SynchronizationType.NONE;
	    		else if(synchType.Equals("stm"))
	    			Parameters.synchronizationType = Parameters.SynchronizationType.STM;
	    		else throw new BenchmarkParametersException("Invalid lock granularity: " + synchType);
	    	}
	    	
            
	    	Type /*Class<? extends SynchMethodInitializer>*/ synchMethodInitializerClass = null;
	    	switch(Parameters.synchronizationType) {
            case Parameters.SynchronizationType.NONE:
	    		synchMethodInitializerClass = ImplParameters.noSynchronizationInitializerClass;
	    		break;
            case Parameters.SynchronizationType.LOCK_COARSE:
	    		synchMethodInitializerClass = ImplParameters.coarseGrainedLockingInitializerClass;
	    		break;
            case Parameters.SynchronizationType.LOCK_MEDIUM:
	    		synchMethodInitializerClass = ImplParameters.mediumGrainedLockingInitializerClass;
	    		break;
            case Parameters.SynchronizationType.LOCK_FINE:
	    		synchMethodInitializerClass = ImplParameters.fineGrainedLockingInitializerClass;
	    		break;
            case Parameters.SynchronizationType.STM:
	    		if(stmInitializerClassName != null) {
	    			try {
	        			synchMethodInitializerClass =
                            (Type/*<? extends SynchMethodInitializer>*/) 
	        					Type.GetType(stmInitializerClassName);    				
	    			}
	    			catch(/*ClassNotFound*/Exception) {
	    			if(synchMethodInitializerClass==null)
                        throw new BenchmarkParametersException("Error instantiating the STM" +
	    						" initializer class");
	    			}
	    		}
	    		else if(ImplParameters.defaultSTMInitializerClass != null) {
	    			synchMethodInitializerClass = ImplParameters.defaultSTMInitializerClass;
	    		}
	    		else throw new BenchmarkParametersException("STM initializer class name not given" +
	    				" and a default class not specified" +
	    				" in ImpParameters.defaultSTMInitializerClass");
	    		break;
	    	}
			try {
			Console.WriteLine(synchMethodInitializerClass);
                //synchMethodInitializer =
                 //   (SynchMethodInitializer) synchMethodInitializerClass.GetConstructor(new Type[0]).Invoke(new Object[0]);
					
				
				if(synchMethodInitializerClass == typeof(CGLockingInitializer))	synchMethodInitializer = new CGLockingInitializer();
				else if(synchMethodInitializerClass == typeof(MGLockingInitializer))	synchMethodInitializer = new MGLockingInitializer();
				//else if(synchMethodInitializerClass == typeof(NoSynchronizationInitializer))	synchMethodInitializer = new NoSynchronizationInitializer();
				else if(synchMethodInitializerClass == typeof(STMInitializer))	synchMethodInitializer = new STMInitializer();
				else throw new Exception("Shouldnt happen");
                    
			}
			catch(Exception e) {
				throw new BenchmarkParametersException("Error instantiating SynchMethod initializer class", e);
			}
			DesignObjFactory.SetInstance(synchMethodInitializer.CreateDesignObjFactory());
			BackendFactory.SetInstance(synchMethodInitializer.CreateBackendFactory());
			OperationExecutorFactory.SetInstance(synchMethodInitializer.CreateOperationExecutorFactory());
		}

	    private void PrintRunTimeParametersInformation() {
	    	PrintSection("Benchmark parameters");
	    	
			
	    	Console.WriteLine("Number of threads: " + Parameters.numThreads);
			Console.WriteLine("Length: " + Parameters.numSeconds + " s");
			Console.WriteLine("Workload: " + Parameters.workloadType);
			Console.WriteLine("Syncronization method: " + Parameters.synchronizationType);
			Console.WriteLine("Long traversals " + (Parameters.longTraversalsEnabled ? "enabled" : "disabled"));
			Console.WriteLine("Structural modification operations " + (Parameters.structureModificationEnabled ? "enabled" : "disabled"));
			//Cause Bartok crash:
			//Console.WriteLine("DesignObjFactory: " + DesignObjFactory.instance.GetType().Name);
			//Console.WriteLine("BackendFactory: " + BackendFactory.instance.GetType().Name);
			//Console.WriteLine("OperationExecutorFactory: " + OperationExecutorFactory.instance.GetType().Name);
			Console.WriteLine("Sequential replay " + (Parameters.sequentialReplayEnabled ? "enabled" : "disabled"));
	    	
			
	    	Console.Write("STM-specific parameters:");
	    	if(Parameters.stmCommandLineParameters == null) Console.Write(" none");
	    	else foreach(String parameter in Parameters.stmCommandLineParameters) 
	    		Console.Write(" " + parameter);
	    	Console.WriteLine("\n");	
	    }
	    
	    private void GenerateOperationCDF() {
	    	double shortTraversalsRatio = (double)Parameters.ShortTraversalsRatio / 100.0, 
	    	operationsRatio = (double)Parameters.OperationsRatio / 100.0;
	    	
	    	double traversalsRatio, structuralModificationsRatio;
	    	if(Parameters.longTraversalsEnabled) traversalsRatio = (double)Parameters.TraversalsRatio / 100.0;
	    	else traversalsRatio = 0;
	    	if(Parameters.structureModificationEnabled)
	    		structuralModificationsRatio = (double)Parameters.StructuralModificationsRatio / 100.0;
	    	else structuralModificationsRatio = 0;

	    	double readOnlyOperationsRatio = (double)Parameters.workloadType.readOnlyOperationsRatio / 100.0,
	    	updateOperationsRatio = 1.0 - readOnlyOperationsRatio;

	    	double sumRatios = traversalsRatio + shortTraversalsRatio + operationsRatio + 
	    		structuralModificationsRatio * updateOperationsRatio;
	    	traversalsRatio /= sumRatios;
	    	shortTraversalsRatio /= sumRatios;
	    	operationsRatio /= sumRatios;
	    	structuralModificationsRatio /= sumRatios;
	    	
	    	OperationId[] operations = OperationId.values();
	    	foreach(OperationId operation in operations) operation.GetType().count++;
	    	
	       	OperationType.TRAVERSAL.probability = 
	       		traversalsRatio * updateOperationsRatio / OperationType.TRAVERSAL.count;
	    	OperationType.TRAVERSAL_RO.probability = 
	    		traversalsRatio * readOnlyOperationsRatio / OperationType.TRAVERSAL_RO.count;
	    	OperationType.SHORT_TRAVERSAL.probability = 
	    		shortTraversalsRatio * updateOperationsRatio / OperationType.SHORT_TRAVERSAL.count;
	    	OperationType.SHORT_TRAVERSAL_RO.probability = 
	    		shortTraversalsRatio * readOnlyOperationsRatio / OperationType.SHORT_TRAVERSAL_RO.count;
	    	OperationType.OPERATION.probability = 
	    		operationsRatio * updateOperationsRatio / OperationType.OPERATION.count;
	    	OperationType.OPERATION_RO.probability = 
	    		operationsRatio * readOnlyOperationsRatio / OperationType.OPERATION_RO.count;
	    	OperationType.STRUCTURAL_MODIFICATION.probability = 
	    		structuralModificationsRatio * updateOperationsRatio / OperationType.STRUCTURAL_MODIFICATION.count;
	    	
	    	Console.WriteLine("Operation ratios [%]:");
	    	foreach(OperationType type in OperationType.values())
	    		Console.WriteLine(AlignText(type.ToString(), 23) + ": " + 
	    				FormatDouble(type.probability * type.count * 100));
	    	Console.WriteLine();
	    	
	    	double[] operationProbabilities = new double[operations.Length];
	    	foreach(OperationId operation in operations) {
	    		double operationProbability = operation.GetType().probability;

                // A operation that has writes that should rarely conflict
                // which we hope may allow stm to beat coarse locking. --Ryan
                if (Parameters.workloadType == WorkloadType.STM_BIASED)
                {
                    if (operation == OperationId.OP9 ) operationProbability = 1.0;
                    else operationProbability = 0.0;

                }

	    		operationProbabilities[operation.ordinal()] = operationProbability;
	    	}
	    	
	    	operationCDF = new double[operations.Length];
	    	double prevProbValue = 0;
	    	for(int opNum = 0; opNum < operations.Length; opNum++) {
	    		operationCDF[opNum] = prevProbValue + operationProbabilities[opNum];
	    		prevProbValue = operationCDF[opNum];
	    	}
	    	operationCDF[operations.Length - 1] = 1.0; // to avoid rounding errors
	    }
	    
	    private void SetupStructures() {
	    	Console.Error.WriteLine("Setup start...");
            setup = new Setup();
	    	
	    	threads = new BenchThread[Parameters.numThreads];
	    	for(short threadNum = 0; threadNum < Parameters.numThreads; threadNum++) {
	    		threads[threadNum] = 
	    			new BenchThread(setup, operationCDF, threadNum);
	    	}
	    	Console.Error.WriteLine("Setup completed.");
	    }

	    private void CheckInvariants(bool initial) {

	    	if(initial) Console.Error.WriteLine("Checking invariants (initial data structure):");
	    	else Console.Error.WriteLine("Checking invariants (final data structure):");

	    	Operation checkInvariantsOperation = new CheckInvariantsOperation(setup, initial);
            OperationExecutorFactory.ExecuteSequentialOperation(checkInvariantsOperation);
	    }
	    
	    private void CreateInitialClone() {
	    	if(! Parameters.sequentialReplayEnabled) return;

	    	Console.Error.WriteLine("Cloning the initial data structure...");
	    	ThreadRandom.Reset();
	   		setupClone = new Setup();
	    	Console.Error.WriteLine("Cloning completed.");

            Console.Error.WriteLine("Checking if the clone is the same as the data structure...");
            StructureComparisonOperation structureComparisonOperation =
                new StructureComparisonOperation(setup, setupClone);
            OperationExecutorFactory.ExecuteSequentialOperation(structureComparisonOperation);
	    	Console.Error.WriteLine("Check OK.");    	
	    }
	    
	    private void Start() {
	    	Console.Error.WriteLine("\nBenchmark started.");
	    	ThreadRandom.StartConcurrentPhase();

            DateTime startTime = DateTime.Now;

            foreach (BenchThread thread in threads) thread.Start();

	    	try {
                Thread.Sleep(Parameters.numSeconds * 1000);
               
	    		foreach(BenchThread thread in threads) thread.StopThread();
                foreach(BenchThread thread in threads) thread.Join();
	    	}
	    	catch(ThreadInterruptedException) {}

            DateTime endTime = DateTime.Now;
            //tick= 100nanos
	    	elapsedTime = (endTime.Ticks - startTime.Ticks)/ 10000000.0;
	    	Console.Error.WriteLine("Benchmark completed.\n");
	    }

	    private void CheckOpacity() {
	    	if(! Parameters.sequentialReplayEnabled) return;
	    	
	    	Console.Error.WriteLine("\nReplaying the execution in a single thread...");

            ArrayList/*<BenchThread.ReplayLogEntry>*/ replayLog = new ArrayList/*<BenchThread.ReplayLogEntry>*/();
            foreach (BenchThread thread in threads)
            {
                //replayLog.AddAll(thread.replayLog);
                replayLog.AddRange(thread.replayLog);
            }
           
            
	    	BenchThread.ReplayLogEntry[] replayLogArray = (BenchThread.ReplayLogEntry[]) replayLog.ToArray/*<BenchThread.ReplayLogEntry>*/(typeof(STMBench7.BenchThread.ReplayLogEntry));
	    	Array.Sort(replayLogArray);
  
	    	replayLog = new ArrayList/*<BenchThread.ReplayLogEntry>*/();
	    	foreach(BenchThread.ReplayLogEntry entry in replayLogArray) replayLog.Add(entry);
	    	replayLogArray = null;
	    	
	    	SequentialReplayThread seqThread = new SequentialReplayThread(setupClone, 
	    			operationCDF, replayLog);
	    	seqThread.Run();
	    	StructureComparisonOperation structureComparisonOperation = 
	    		new StructureComparisonOperation(setup, setupClone);

            OperationExecutorFactory.ExecuteSequentialOperation(structureComparisonOperation);
	    	Console.Error.WriteLine("\nOpacity ensured.\n");
	    }
	    
	    private void ShowTTCHistograms() {
	    	if(! printTTCHistograms) return;
	    	
	    	PrintSection("TTC histograms");
	    	
	    	OperationId[] operations = OperationId.values();
	    	foreach(OperationId operation in operations) {
	    		Console.Write("TTC histogram for " + operation + ":");
	    		
	    		for(int ttc = 0; ttc <= Parameters.MAX_LOW_TTC; ttc++) {
	        		int count = 0;
	        		foreach(BenchThread thread in threads) 
	        			count += thread.operationsTTC[operation.ordinal(),ttc];
	        		
	        		Console.Write(" " + ttc + "," + count);
	    		}
	    		
	    		for(int logTtcIndex = 0; logTtcIndex < Parameters.HIGH_TTC_ENTRIES; logTtcIndex++) {
	    			int count = 0;
	    			foreach(BenchThread thread in threads)
	    				count += thread.operationsHighTTCLog[operation.ordinal(),logTtcIndex];
	    			
	    			int ttc = LogTtcIndex2Ttc(logTtcIndex);
	    			Console.Write(" " + ttc + "," + count);
	    		}
	    		
	    		Console.WriteLine();
	    	}
	    	Console.WriteLine();
	    }
	    
	    private int LogTtcIndex2Ttc(double logTtcIndex) {
	    	return (int)((Parameters.MAX_LOW_TTC + 1) * Math.Pow(Parameters.HIGH_TTC_LOG_BASE, logTtcIndex));
	    }
	    
	    private void ShowStats() {
	    	PrintSection("Detailed results");
	    	
	    	OperationId[] operations = OperationId.values();
	    	foreach(OperationId operation in operations) {
	    		Console.Write("Operation " + AlignText(operation.ToString(), 4) + ":   ");
	    		
	    		int opNumber = operation.ordinal();
	    		int successful = 0, failed = 0, maxttc = 0;
	    		foreach(BenchThread thread in threads) {
	    			successful += thread.successfulOperations[opNumber];
	    			failed += thread.failedOperations[opNumber];
	    			maxttc = Math.Max(maxttc, ComputeMaxThreadTTC(thread, opNumber));
	    		}
	    		
	    		Console.WriteLine("successful = " + successful + "\tmaxttc = " + maxttc + "\tfailed = " + failed);
	    		
	    		OperationType operationType = operation.GetType();
	    		operationType.successfulOperations += successful;
	    		operationType.failedOperations += failed;
	    		operationType.maxttc = Math.Max(operationType.maxttc, maxttc);
	    	}
	    	Console.WriteLine();

	    	PrintSection("Sample errors (operation ratios [%])");

	    	int totalSuccessful = 0, totalFailed = 0;
	    	foreach(OperationType type in OperationType.values()) {
	    		totalSuccessful += type.successfulOperations;
	    		totalFailed += type.failedOperations;
	    	}

	    	double totalError = 0, totalTError = 0;
	    	foreach(OperationType type in OperationType.values()) {
	    		double expectedRatio = type.probability * type.count * 100.0;
	    		double realRatio = (double)type.successfulOperations / (double)totalSuccessful * 100.0; 
	    		double error = Math.Abs(realRatio - expectedRatio);
	    		double tRealRatio = (double)(type.successfulOperations + type.failedOperations) /
	    			(double)(totalSuccessful + totalFailed) * 100.0;
	    		double tError = Math.Abs(tRealRatio - expectedRatio);
	    		Console.WriteLine(AlignText(type.ToString(), 23) + ":  " + 
	    				"expected = " + FormatDouble(expectedRatio) +
	    				"\tsuccessful = " + FormatDouble(realRatio) +
	    				"\terror = " + FormatDouble(error) +
	    				"\t(total = " + FormatDouble(tRealRatio) +
	    				"\terror = " + FormatDouble(tError) + ")");
	    		totalError += error;
	    		totalTError += tError;
	    	}
	    	Console.WriteLine();
	    	
	    	PrintSection("Summary results");
	    	
	    	int total = totalSuccessful + totalFailed;
	    	foreach(OperationType type in OperationType.values()) {
	    		int totalTypeOperations = type.successfulOperations + type.failedOperations;
	    		Console.WriteLine(AlignText(type.ToString(), 23) + ":  " +
	    				"successful = " + type.successfulOperations + 
	    				"\tmaxttc = " + type.maxttc +
	    				"\tfailed = " + type.failedOperations +
	    				"\ttotal = " + totalTypeOperations);
	    	}
	    	Console.WriteLine();
	    	
	    	Console.WriteLine("Total sample error: " + FormatDouble(totalError) + "%" +
	    			" (" + FormatDouble(totalTError) + "% including failed)");
	    	Console.WriteLine("Total throughput: " + FormatDouble( (double)totalSuccessful / elapsedTime ) + " op/s" +
	    			" (" + FormatDouble( (double)total / elapsedTime ) + " op/s including failed)");
	    	Console.WriteLine("Elapsed time: " + FormatDouble(elapsedTime) + " s");
	    }

	    private int ComputeMaxThreadTTC(BenchThread thread, int opNumber) {
	    	for(int logTtcIndex = Parameters.HIGH_TTC_ENTRIES - 1; logTtcIndex >= 0; logTtcIndex--) {
	    		if(thread.operationsHighTTCLog[opNumber,logTtcIndex] > 0)
	    			return LogTtcIndex2Ttc(logTtcIndex);
	    	}
	    	
	    	for(int ttc = Parameters.MAX_LOW_TTC; ttc >= 0; ttc--) {
	    		if(thread.operationsTTC[opNumber,ttc] > 0) return ttc;
	    	}
	    	
	    	return 0; // operation never completed with success
	    }
	    
	    private void PrintSyntax() {
	    	String syntax = 
	    		"Syntax:\n" +
	    		"java stmbench7.Benchmark [options] [-- stm-specific options]\n\n" +
	    		"Options:\n" +
	    		"\t-t numThreads -- set the number of threads (default: 1)\n" +
	    		"\t-l length     -- set the length of the benchmark, in seconds (default: 10)\n" +
	    		"\t-w r|rw|w     -- set the workload: r = read-dominated, w = write-dominated\n" +
	    		"\t                                   rw = read-write (default: read-dominated)\n" +
	    		"\t-g coarse|medium|fine|none|stm -- set synchronization method (default: coarse)\n" +
	    		"\t-s stmInitializerClass         -- set STM initializer class (default: none)\n" +
	    		"\t--no-traversals  -- do not use long traversals\n" +
	    		"\t--no-sms         -- do not use structural modification operations\n" +
	    		"\t--seq-replay 	-- replay the execution in a single thread\n" +
	    		"\t                    (checks for opacity violations)\n" +
	    		"\t--ttc-histograms -- print TTC histograms to stdout\n\n" +
	    		"Note: the benchmark needs a lot of lot of memory, so the -Xmx option of Java\n" +
	    		"might be necessary.";
			Console.Error.WriteLine(syntax);
	    }
	    
	    private void PrintSection(String title) {
	    	PrintLine('-');
	    	Console.WriteLine(title);
	    	PrintLine('-');
	    }
	    
	    private void PrintLine(char ch) {
	    	StringBuilder line = new StringBuilder(79);
            for (int i = 0; i < 79; i++)
            {
                line.Append(ch);
            }
	    	Console.WriteLine(line);
	    }
	    
	    private String AlignText(String text, int width) {
	    	int textLen = text.Length;
	    	int padding = width - textLen;
	    	if(padding < 0) throw new RuntimeError("alignText: width too small!");
	    	
	    	StringBuilder output = new StringBuilder(width);
            for (int i = 0; i < padding; i++)
            {
                output.Append(' ');
            }
	    	output.Append(text);
	    	
	    	return output.ToString();
	    }
	    
	    private String FormatDouble(double number) {
	    	//Formatter formatter = new Formatter();
	    	//formatter.format("%3.2f", number);
	    	//return formatter.ToString();
            return String.Format("{0:###.00}", number);
	    }
	}

}
