﻿using System;
 
 
using System.Text;

using STMBench7.Annotations;
using STMBench7.Core;
using STMBench7.Compatability;
using System.Threading;

namespace STMBench7
{
	/**
	 * This is a central repository for thread-local random
	 * number generators. No other class should create an instance
	 * of class Random, but should use the methods in ThreadRandom
	 * instead. This way we can centrally control the (un)determinism
	 * of the benchmark and the implementation of a random number
	 * generator used.
	 */
	[Immutable]
	public class ThreadRandom {

		private /*static*/ enum Phase {
            INIT,
            CONCURRENT,
            SEQUENTIAL_REPLAY
		}
		
		private static Phase phase = Phase.INIT;
        private static CannedRandom initRandom = new CannedRandom(3);
		private static short currentVirtualThreadNumber;
        private static CannedRandom[] virtualRandom;
		
		protected ThreadRandom() { }

        static GenomeHashtable threadLocalRandoms = new GenomeHashtable();

        static CannedRandom ThreadLocalRandom
        {
            get
            {
                lock (threadLocalRandoms)
                {
                    if (!threadLocalRandoms.Contains(Thread.CurrentThread))
                    {
                        threadLocalRandoms.Add(Thread.CurrentThread, new CannedRandom(3));
                    }
                    return (CannedRandom)threadLocalRandoms.Get(Thread.CurrentThread);
                }
            }
        }
        
		
		public static int NextInt() {
            return GetCurrentRandom().NextInt();
		}
		
		public static int NextInt(int n) {
            return GetCurrentRandom().NextInt(n);
		}
		
		public static double NextDouble() {
			return GetCurrentRandom().NextDouble();
		}

		public static void Reset() {
			if(phase != Phase.INIT) 
				throw new RuntimeError("Cannot reset ThreadRandom after the initialization phase");
            initRandom = new CannedRandom(3);
		}
		
		public static void StartConcurrentPhase() {
			phase = Phase.CONCURRENT;
		}
		
		public static void StartSequentialReplayPhase() {
			phase = Phase.SEQUENTIAL_REPLAY;
            virtualRandom = new CannedRandom[Parameters.numThreads];
			for(int n = 0; n < Parameters.numThreads; n++)
                virtualRandom[n] = new CannedRandom(3);
		}

		public static void SetVirtualThreadNumber(short threadNum) {
			currentVirtualThreadNumber = threadNum;
		}

        private static CannedRandom GetCurrentRandom()
        {
			switch(phase) { 
			case Phase.INIT: return initRandom;
            case Phase.CONCURRENT: return ThreadLocalRandom;
			case Phase.SEQUENTIAL_REPLAY: return virtualRandom[currentVirtualThreadNumber];
			default: return null;
			}
		}
	}

}
